// astretch.c
// makes a new RGB image by  automatic histogram stretching
// heavily, and in a dirty(?) way, reuses code from resample.c 
// and on Contrast Autostretch 1.04 --- image filter plug-in for The Gimp image
// manipulation program,  by Federico Mena Quintero 
// <federico@nuclecu.unam.mx, quartic@polloux.fciencias.unam.mx>
// sorry, and thanks!
// Thu Mar 31 15:01:15 MET 2005 Peter A. Cejchan <cej@gli.cas.cz | pac7@post.cz>

// License: GPL


/* This code does an automatic contrast stretch.  For each
   channel in the image, it finds the minimum and maximum values... it
   uses those values to stretch the individual histograms to the full
   contrast range.  For some images it may do just what you want; for
   others it may be total crap :) 
*/




#include <u.h>
#include <libc.h>
#include <draw.h>
#include <memdraw.h>

#define ALPHA 3

// VERSION:
char version[10] = "1.02";



void
usage(void)
{
	fprint(2, "usage:\n\tastretch [imagefile]\n");
	fprint(2, "\n\tjpg -3 testimg.jpg | astretch  | page  \n");
	fprint(2, "\n\tastretch  rose.img  | togif | gif\n");
	exits("usage");
}

void
help(void)
{
	fprint(2, "astretch-%s\n",version);
	fprint(2, "Contrast Autostretch\n");
	usage();
}


static void
do_astretch(uchar **oscan, uchar **nscan, int xsize, int ysize, int channels)
{
	uchar *src, *dest, old, new;
	uchar  min[3], max[3];  /* Minimums and maximums */
	uchar  range;
	uchar  lut[256][3];
	int            x1, y1, x2, y2;
	long          i, x, y, off;
	int            k;


	switch(channels){
	case 3:
	case 4: 
		break;
	default:
		sysfatal("can't handle  %d-channel images, need 24(32)-bpp depth", channels);	
	}

	/* Get minimum and maximum values for each channel */

	min[0] = min[1] = min[2] = 255;
	max[0] = max[1] = max[2] = 0;

	for (x = 0; x < xsize; x++) {
		for (k = 0; k < channels; k++){
			off=channels*x+k;
			for(y=0; y<ysize; y++){  // for every point in scanline
				old =  oscan[y][off];
					if (old < min[k])
						min[k] = old;
					if (old > max[k])
						max[k] = old;
			}
		}

	} 

	/* Calculate LUTs with stretched contrast */

	for (k = 0; k < channels; k++) {
		range = max[k] - min[k];

		if (range != 0)
			for (x = min[k]; x <= max[k]; x++)
				lut[x][k] = 255 * (x - min[k]) / range;
		else
			lut[min[k]][k] = min[k];
	} /* for */

	/* Now substitute pixel vales */

	for (x = 0; x < xsize; x++) {
		for (k = 0; k < channels; k++) {
			off=channels*x+k;
			for (y = 0; y < ysize; y++) {
				old =  oscan[y][off];
				if (k != ALPHA)
					nscan[y][off]  = lut[old][k];
				else
					nscan[y][off] = old;
				 
			}
		}
	} 
} /* do_astretch */




int
max(int a, int b)
{
	if(a > b)
		return a;
	return b;
}


Memimage*
mknewimg(int xsize, int ysize, Memimage *m)
{
	int i, j, bpl, nchan;
	Memimage *new;
	uchar **oscan, **nscan;

	new = allocmemimage(Rect(0, 0, xsize, ysize), m->chan);
	if(new == nil)
		sysfatal("can't allocate new image: %r");

	oscan = malloc(Dy(m->r)*sizeof(uchar*));
	nscan = malloc(max(ysize, Dy(m->r))*sizeof(uchar*));
	if(oscan == nil || nscan == nil)
		sysfatal("can't allocate: %r");

	/* unload original image into scan lines */
	bpl = bytesperline(m->r, m->depth);
	for(i=0; i<Dy(m->r); i++){
		oscan[i] = malloc(bpl);
		if(oscan[i] == nil)
			sysfatal("can't allocate: %r");
		j = unloadmemimage(m, Rect(m->r.min.x, m->r.min.y+i, m->r.max.x, m->r.min.y+i+1), oscan[i], bpl);
		if(j != bpl)
			sysfatal("unloadmemimage");
	}

	/* allocate scan lines for destination. we do y first, so need at least Dy(m->r) lines */
	bpl = bytesperline(Rect(0, 0, xsize, Dy(m->r)), m->depth);
	for(i=0; i<max(ysize, Dy(m->r)); i++){
		nscan[i] = malloc(bpl);
		if(nscan[i] == nil)
			sysfatal("can't allocate: %r");
	}

	nchan = m->depth/8;

	do_astretch(oscan, nscan, xsize, ysize, nchan);



	/* pack data into destination */
	bpl = bytesperline(new->r, m->depth);
	for(i=0; i<ysize; i++){
		j = loadmemimage(new, Rect(0, i, xsize, i+1), nscan[i], bpl);
		if(j != bpl)
			sysfatal("loadmemimage: %r");
	}
	return new;
}


void
main(int argc, char *argv[])
{
	int i, fd, xsize, ysize, dummy;
	Rectangle rparam;
	Memimage *m, *new, *t1, *t2;
	char *file;
	ulong tchan;
	char tmp[100];
	double v;

	memimageinit();
	memset(&rparam, 0, sizeof rparam);
	xsize = ysize = 0;


	file = "<stdin>";
	fd = 0;



	ARGBEGIN{
	case 'h':
		help();
		break;
	}ARGEND

	switch (argc) {
	case 1:
		file = argv[0];
		fd = open(file, OREAD);
		if(fd < 0)
			sysfatal("can't open %s: %r", file);
	case 0:
		break;
	default:
		usage();
	}

	m = readmemimage(fd);
	if(m == nil)
		sysfatal("can't read %s: %r", file);

	xsize = Dx(m->r);
	ysize = Dy(m->r);

	new = nil;
	switch(m->chan){

	case RGB24:
	case RGBA32:
	case ARGB32:
	case XRGB32:
		new = mknewimg(xsize, ysize, m);
		break;

	case GREY1:
	case GREY2:
	case GREY4:
	case GREY8:
	case CMAP8:
	case RGB15:
	case RGB16:
		tchan = RGB24;
		goto Convert;

	Convert:
		/* use library to convert to byte-per-chan form, then convert back */
		t1 = allocmemimage(m->r, tchan);
		if(t1 == nil)
			sysfatal("can't allocate temporary image: %r");
		memimagedraw(t1, t1->r, m, m->r.min, nil, ZP, S);
		t2 = mknewimg(xsize, ysize, t1);
		freememimage(t1);
		new = allocmemimage(Rect(0, 0, xsize, ysize), m->chan);
		if(new == nil)
			sysfatal("can't allocate new image: %r");
		/* should do error diffusion here */
		memimagedraw(new, new->r, t2, t2->r.min, nil, ZP, S);
		freememimage(t2);
		break;

	default:
		sysfatal("can't handle channel type %s", chantostr(tmp, m->chan));
	}

	assert(new);
	if(writememimage(1, new) < 0)
		sysfatal("write error on output: %r");

	exits(nil);
}
